@@ -1,21 +1,21 @@
-mod allocator;
+pub mod allocator;
 pub mod chia_dialect;
 mod core_ops;
-mod cost;
+pub mod cost;
 mod dialect;
 mod err_utils;
-mod f_table;
+pub mod f_table;
 mod gen;
 mod int_to_bytes;
-mod more_ops;
-mod node;
+pub mod more_ops;
+pub mod node;
 mod number;
 mod op_utils;
-mod operator_handler;
+pub mod operator_handler;
 #[cfg(not(any(test, target_family = "wasm")))]
 mod py;
-mod reduction;
-mod run_program;
+pub mod reduction;
+pub mod run_program;
 mod serialize;
 mod sha2;
 
