@@ -97,11 +97,18 @@ fn convert_spend(
     }
 
     let mut new_coins = Vec::<PyConditionWithArgs>::new();
-    for (ph, amount) in spend_cond.create_coin {
-        new_coins.push(PyConditionWithArgs {
+    for new_coin in spend_cond.create_coin {
+        let mut cwa = PyConditionWithArgs {
             opcode: CREATE_COIN,
-            vars: vec![PyBytes::new(py, &ph).into(), int_to_pybytes(py, amount)],
-        });
+            vars: vec![
+                PyBytes::new(py, &new_coin.puzzle_hash).into(),
+                int_to_pybytes(py, new_coin.amount),
+            ],
+        };
+        if new_coin.hint != a.null() {
+            cwa.vars.push(node_to_pybytes(py, a, new_coin.hint));
+        }
+        new_coins.push(cwa);
     }
     if !new_coins.is_empty() {
         ordered.insert(CREATE_COIN, new_coins);
